#pragma once

#include "BlockPreconditioner.hpp"
#include <amdis/common/Literals.hpp>

namespace AMDiS
{
  template <class Matrix, class Vector>
  class MTLPfcPrecon
      : public BlockPreconditioner<Matrix, Vector>
  {
  public:
    using Self = MTLPfcPrecon;
    using Super = BlockPreconditioner<Matrix, Vector>;
    using PreconBase = typename Super::PreconBase;

    using MTLMatrix = BaseMatrix_t<Matrix>;
    using MTLVector = Vector;
    using LinearSolverType = LinearSolverInterface<MTLMatrix, MTLVector>;

  public:
    struct Creator : CreatorInterface<PreconBase>
    {
      virtual std::shared_ptr<PreconBase> create() override
      {
        precon = std::make_shared<Self>();
        return precon;
      }

      std::shared_ptr<Self> precon = NULL;
    };

  public:
    MTLPfcPrecon()
      : Super()
    {
      std::string solverNameM = "cg",
                  solverNameMpL = "cg",
                  solverNameMpL2 = "cg";
      Parameters::get("precon_pfc->M->solver", solverNameM);
      Parameters::get("precon_pfc->MpL->solver", solverNameMpL);
      Parameters::get("precon_pfc->MpL2->solver", solverNameMpL2);

      CreatorInterfaceName<LinearSolverType>* creator;

      creator = named( CreatorMap<LinearSolverType>::getCreator(solverNameM, "precon_pfc->M->solver") );
      solverM = creator->create("precon_pfc->M");

      creator = named( CreatorMap<LinearSolverType>::getCreator(solverNameMpL, "precon_pfc->MpL->solver") );
      solverMpL = creator->create("precon_pfc->MpL");

      creator = named( CreatorMap<LinearSolverType>::getCreator(solverNameMpL2, "precon_pfc->MpL2->solver") );
      solverMpL2 = creator->create("precon_pfc->MpL2");
    }

    virtual ~MTLPfcPrecon()
    {
      exit();
    }

    void setData(double* tau_, double M0_ = 1.0)
    {
      tau = tau_;
      M0 = M0_;
    }

    // implementation in MTLPfcPrecon.inc.hpp
    virtual void init(Matrix const& A) override;

    virtual void exit() override {}

    // implementation in MTLPfcPrecon.inc.hpp
    virtual void solve(MTLVector const& b, MTLVector& x) const override;


    template <typename VectorX, typename VectorB>
    void solveM(VectorX& x, const VectorB& b) const
    {
      SolverInfo solverInfo("precon_pfc->M");
      solverM->solve(getM(), x, b, solverInfo);
    }

    template <typename VectorX, typename VectorB>
    void solveMpL(VectorX& x, const VectorB& b) const
    {
      SolverInfo solverInfo("precon_pfc->MpL");
      solverMpL->solve(MpL, x, b, solverInfo);
    }

    template <typename VectorX, typename VectorB>
    void solveMpL2(VectorX& x, const VectorB& b) const
    {
      SolverInfo solverInfo("precon_pfc->MpL2");
      solverMpL2->solve(MpL2, x, b, solverInfo);
    }

    MTLMatrix const& getM()  const { return matM  ? *matM  : (*Super::A)(_2, _2); } // < u, v >
    MTLMatrix const& getL0() const { return matL0 ? *matL0 : (*Super::A)(_1, _0); } // < M0*tau*grad(u), grad(v) >
    MTLMatrix const& getL()  const { return matL  ? *matL  : (*Super::A)(_2, _1); } // < grad(u), grad(v) >

    double getTau() const { return *tau; }

  protected:
    double* tau = NULL;
    double M0 = 1.0;

    MTLMatrix* matM = NULL;
    MTLMatrix* matL = NULL;
    MTLMatrix* matL0 = NULL;

    MTLMatrix MpL;
    MTLMatrix MpL2;

    mutable MTLVector y0;
    mutable MTLVector y1;
    mutable MTLVector tmp;

    std::shared_ptr<LinearSolverType> solverM;
    std::shared_ptr<LinearSolverType> solverMpL;
    std::shared_ptr<LinearSolverType> solverMpL2;
  };

} // end namespace AMDiS

#include "MTLPfcPrecon.inc.hpp"
